
#ifndef __TYPELIST_ADDITIONS_H__
#define __TYPELIST_ADDITIONS_H__

#include "Typelist.h"

namespace Loki
{
	namespace TL
	{
		//////////////////////////////////////////////////////////////
		// Subtract													//
		//////////////////////////////////////////////////////////////
		template <class TListMinuend, class TListSubtrahed > struct Subtract;

		template <class TListSubtrahed> 
		struct Subtract<NullType, TListSubtrahed>
		{
			typedef NullType Result;
		};

		template <class Head, class Tail> 
		struct Subtract<Typelist<Head, Tail>, NullType>
		{
			typedef Typelist<Head, Tail> Result;
		};

		template <class Head1, class Tail1, class Head2, class Tail2> 
		struct Subtract<Typelist<Head1, Tail1>, Typelist<Head2, Tail2> >
		{
			private:
				typedef typename TL::Erase<Typelist<Head1, Tail1>, Head2>::Result L1;
			public:
				typedef typename Subtract<L1, Tail2>::Result Result;
		};

	
		//////////////////////////////////////////////////////////////
		// Intersection												//
		//////////////////////////////////////////////////////////////
		template <class TFirstList, class TSecondList > struct Intersection;
		
		template <class TFirstList>
		struct Intersection<NullType, TFirstList>
		{
			typedef NullType Result;
		};
	
		template <class Head, class Tail, class TSecondList >
		struct Intersection<Typelist<Head, Tail>, TSecondList>
		{
			private:
				enum { temp = IndexOf<TSecondList, Head>::value };
				typedef typename Select<temp == -1, NullType, Head >::Result L1;
			public:		
		 		typedef typename Append<typename Intersection<Tail, TSecondList>::Result, L1 >::Result Result;		
		};		
		
		//////////////////////////////////////////////////////////////
		// EraseIf													//
		//////////////////////////////////////////////////////////////
		template <class TList, template<class> class TRemoveCondition > struct EraseIf;
		
		template <template<class> class TRemoveCondition>
		struct EraseIf<NullType, TRemoveCondition>
		{
			typedef NullType Result;
		};
	
		template <class Head, class Tail, template<class> class TRemoveCondition >
		struct EraseIf<Typelist<Head, Tail>, TRemoveCondition>
		{
			public:		
				typedef typename Select< (TRemoveCondition<Head>::value), 
								typename EraseIf<Tail, TRemoveCondition>::Result , 
								Typelist <Head, typename EraseIf<Tail, TRemoveCondition>::Result > >::Result Result;
		};
		
		//////////////////////////////////////////////////////////////
		// FindIf													//
		//////////////////////////////////////////////////////////////
		template <class TList, template<class> class TStopCondition > struct FindIf;

		template <template<class> class TStopCondition>
		struct FindIf<NullType, TStopCondition>
		{
			typedef NullType Result;
		};
	
		template <class Head, class Tail, template<class> class TStopCondition >
		struct FindIf<Typelist<Head, Tail>, TStopCondition>
		{
			public:		
				typedef typename Select< (TStopCondition<Head>::value), Head , 
								typename FindIf<Tail, TStopCondition>::Result >::Result Result;
		};

		//////////////////////////////////////////////////////////////
		// SubList (From Head to cnt or NullType)													//
		//////////////////////////////////////////////////////////////

		template <class TList, unsigned long cnt> struct SubList;

		template <unsigned long cnt> 
		struct SubList<NullType, cnt>
		{
			typedef NullType Result;
		};

		template <class Head, class Tail> 
		struct SubList<Typelist<Head, Tail>, 0>
		{
			typedef NullType Result;
		};

		template <class Head, class Tail, unsigned long cnt> 
		struct SubList <Typelist<Head, Tail>, cnt>
		{
			typedef Typelist <Head, typename SubList<Tail, cnt - 1>::Result> Result;
		};

	}
}


#endif // __TYPELIST_ADDITIONS_H__